/*
 * autotimer.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <vdr/plugin.h>
#include <vdr/timers.h>

#define FULLSCAN_INTERVAL   (24*60*60)  // 24 hours
#define QUICKSCAN_INTERVAL     (10*60)  // 10 min

#include "rules.h"

static void CreateTimer(cEvent *event)
{
  if(event) {
    cTimer *timer = new cTimer(event);
    cTimer *t = Timers.GetTimer(timer);
    cString date = event->GetDateString();
    cString time = event->GetTimeString();
    const char *title  = event->Title();
    if (t) {
      isyslog("autotimer: Timer for %s %s %s already exists", *date, *time, title);
      delete timer;
    } else if(!timer) {
      esyslog("autotimer: Error creating timer for %s %s %s", *date, *time, title);
    } else {
      //
      // TODO: keep record and check if program has already been recorded (re-transmission) ...
      //
      Timers.Add(timer);
      timer->Matches();
      Timers.SetModified();
      Timers.Save();
      isyslog("autotimer: Created timer %s %s %s", *date, *time, title);
    }
  }
}

static int64_t ScanEpg(bool Quick)
{
  time_t now = time(NULL);
  time_t FirstEnd = now + QUICKSCAN_INTERVAL;
  int ch = 1;
  
  cString SpecFile = AddDirectory(cPlugin::ConfigDirectory(), "autotimer.spec");
  cRule *Rule = cRule::RuleFromFile(*SpecFile);
  if(!Rule) {
    esyslog("autotimer: Can't load and parse %s", *SpecFile);
    return Quick ? QUICKSCAN_INTERVAL : FULLSCAN_INTERVAL;
  }

  cSchedulesLock schedulesLock;
  const cSchedules *schedules = cSchedules::Schedules(schedulesLock);

  if(schedules) {
    cChannel *Channel = Channels.GetByNumber(ch);
    while(Channel) {
      const cSchedule *schedule = schedules->GetSchedule(Channel->GetChannelID());
      if(schedule && schedule->Events()) {
	int n = 0;
	for (cEvent *p = schedule->Events()->First(); p && (!Quick || n < 5);
	     p = schedule->Events()->Next(p), n++) {
	  if(Rule->Matches(p))
	    CreateTimer(p);
	  if((n==4 || !schedule->Events()->Next(p)) && p->EndTime() < FirstEnd) {
	    if(p->EndTime() > now)
	      FirstEnd = p->EndTime();
	  }
	}
      }
      Channel = Channels.GetByNumber(++ch);
    }   
  }
  return FirstEnd - time(NULL);
}


//--- cPluginAutotimer -------------------------------------------

static char *TimeToStr(time_t sec)
{
  static char buffer[64];
  struct tm tm_r;

  localtime_r(&sec, &tm_r);
  strftime(buffer, 63, "%Y-%m-%d %T", &tm_r);

  return buffer;
}

static const char *VERSION        = "0.1.0";
static const char *DESCRIPTION    = "Enter description for 'autotimer' plugin";
static const char *MAINMENUENTRY  = NULL;//"Autotimer EPG scan";

class cPluginAutotimer : public cPlugin {
private:
  // Add any member variables or functions you may need here.

  int64_t NextFullScan;
  int64_t NextQuickScan;

public:
  cPluginAutotimer(void);
  virtual ~cPluginAutotimer();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return DESCRIPTION; }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Initialize(void);
  virtual bool Start(void);
  virtual void Stop(void);
  virtual void Housekeeping(void);
  virtual void MainThreadHook(void);
  virtual cString Active(void);
  virtual const char *MainMenuEntry(void) { return MAINMENUENTRY; }
  virtual cOsdObject *MainMenuAction(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  virtual bool Service(const char *Id, void *Data = NULL);
  virtual const char **SVDRPHelpPages(void);
  virtual cString SVDRPCommand(const char *Command, const char *Option, int &ReplyCode);

  //void ScanEpg();
  };

cPluginAutotimer::cPluginAutotimer(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!

  NextFullScan  = time(NULL) + FULLSCAN_INTERVAL;
  NextQuickScan = time(NULL) + QUICKSCAN_INTERVAL;
}

cPluginAutotimer::~cPluginAutotimer()
{
  // Clean up after yourself!
}

const char *cPluginAutotimer::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return NULL;
}

bool cPluginAutotimer::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginAutotimer::Initialize(void)
{
  // Initialize any background activities the plugin shall perform.
  return true;
}

bool cPluginAutotimer::Start(void)
{
  // Start any background activities the plugin shall perform.
  NextFullScan  = NextQuickScan = time(NULL) + 60;
  return true;
}

void cPluginAutotimer::Stop(void)
{
  // Stop any background activities the plugin shall perform.
}

void cPluginAutotimer::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.

  // If this takes much time, do in slices
  time_t now = time(NULL);
  if(NextFullScan <= now) {
    int64_t now_ms = cTimeMs::Now();
    isyslog("AutoTimer: performing full scan (%d seconds late)",
	    (int)(now - NextFullScan));
    ScanEpg(false);
    NextFullScan = now + FULLSCAN_INTERVAL;
    isyslog("AutoTimer: Full scan done (%d ms). Next full scan at %s",
	    (int)(cTimeMs::Now() - now_ms), TimeToStr(NextFullScan));
  }
}

void cPluginAutotimer::MainThreadHook(void)
{
  // Perform actions in the context of the main program thread.
  // WARNING: Use with great care - see PLUGINS.html!

  time_t now = time(NULL);
  if(NextQuickScan <= now) {
    int64_t now_ms = cTimeMs::Now();
    isyslog("AutoTimer: performing quick scan (%d seconds late)",
	    (int)(now - NextQuickScan));
    NextQuickScan = now + ScanEpg(true);
    if(NextQuickScan > (now + QUICKSCAN_INTERVAL))
      NextQuickScan = now + QUICKSCAN_INTERVAL;
    isyslog("AutoTimer: Quick scan done (%d ms). Next quick scan at %s",
	    (int)(cTimeMs::Now() - now_ms), TimeToStr(NextQuickScan));
  }
}

cString cPluginAutotimer::Active(void)
{
  // Return a message string if shutdown should be postponed
  return NULL;
}

cOsdObject *cPluginAutotimer::MainMenuAction(void)
{
  // Perform the action when selected from the main VDR menu.

  int64_t now = time(NULL);
  NextFullScan = now;
  Housekeeping();

  return NULL;
}

cMenuSetupPage *cPluginAutotimer::SetupMenu(void)
{
  // Return a setup menu in case the plugin supports one.

  int64_t now = time(NULL);
  NextFullScan = now;
  Housekeeping();

  return NULL;
}

bool cPluginAutotimer::SetupParse(const char *Name, const char *Value)
{
  // Parse your own setup parameters and store their values.
  return false;
}

bool cPluginAutotimer::Service(const char *Id, void *Data)
{
  // Handle custom service requests from other plugins
  return false;
}

const char **cPluginAutotimer::SVDRPHelpPages(void)
{
  // Return help text for SVDRP commands this plugin implements
  return NULL;
}

cString cPluginAutotimer::SVDRPCommand(const char *Command, const char *Option, int &ReplyCode)
{
  // Process SVDRP commands this plugin implements
  return NULL;
}

VDRPLUGINCREATOR(cPluginAutotimer); // Don't touch this!
