/*
 * rules.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <vdr/epg.h>

#include "rules.h"

typedef cList<cRule> cRules;


static char *Ident(int ident)
{
  static char str[1024];
  str[ident] = 0;
  while(ident) 
    str[--ident] = ' ';
  return str;
}

//--- cRuleTitleSubstring ---------------------------------------

class cRuleTitleSubstring : public cRule 
{
  protected:
    cString m_Str;
  public:
    cRuleTitleSubstring(const char *String) : m_Str(String) {}
    virtual bool Matches(cEvent *Event) const {
      return strstr(Event->Title(), *m_Str);
    }
    virtual cString ToText(int ident) const { 
      return cString::sprintf("%sTITLE CONTAINS \"%s\"", Ident(ident), *m_Str);
    }
};

//--- cRuleSubstring --------------------------------------------

class cRuleSubstring : public cRule 
{
  protected:
    cString m_Str;
  public:
    cRuleSubstring(const char *String) : m_Str(String) {}
    virtual bool Matches(cEvent *Event) const {
      return
	(Event->Title() &&       strstr(Event->Title(),       *m_Str)) ||
	(Event->ShortText() &&   strstr(Event->ShortText(),   *m_Str)) ||
	(Event->Description() && strstr(Event->Description(), *m_Str));
    }
    virtual cString ToText(int ident) const { 
      return cString::sprintf("%sINFO CONTAINS \"%s\"", Ident(ident), *m_Str);
    }
};

//--- cRuleTitle ------------------------------------------------

class cRuleTitle : public cRule 
{
  protected:
    cString m_Str;
  public:
    cRuleTitle(const char *String) : m_Str(String) {}
    virtual bool Matches(cEvent *Event) const {
      return !strcasecmp(Event->Title(), *m_Str);
    }
    virtual cString ToText(int ident) const { 
      return cString::sprintf("%sTITLE IS \"%s\"", Ident(ident), *m_Str);
    }
};

//--- cRuleChannelRange -----------------------------------------

#include <vdr/channels.h>

class cRuleChannelRange : public cRule 
{
  protected:
    int m_Start;
    int m_End;
  public:
    cRuleChannelRange(int Start, int End) : m_Start(Start), m_End(End) {}
    virtual bool Matches(cEvent *Event) const {
      cChannel *ch = Channels.GetByChannelID(Event->ChannelID());
      if(ch)
	if(ch->Number() < m_Start || ch->Number() > m_End)
	  return false;
      return true;
    }
    virtual cString ToText(int ident) const { 
      if(m_Start==m_End)
	return cString::sprintf("%sCHANNELNUMBER IS %d", Ident(ident), m_Start);      
      return cString::sprintf("%sCHANNELNUMBER IS %d-%d", Ident(ident), m_Start, m_End);
    }
};

//--- cRuleNot ---------------------------------------------------

class cRuleNot : public cRule 
{
  private:
    cRuleNot(cRuleNot&);
  protected:
    cRule *m_Rule;
  public:
    cRuleNot(cRule *Rule) : m_Rule(Rule) {}
    virtual ~cRuleNot() { delete m_Rule; }
    virtual bool Matches(cEvent *Event) const {
      return !m_Rule->Matches(Event);
    }
    virtual cString ToText(int ident) const { 
      return cString::sprintf("%sNOT {%s}", Ident(ident), *m_Rule->ToText());
    }
};

//--- cRuleAnd --------------------------------------------------

#include <stdarg.h>

class cRuleAnd : public cRule 
{
  protected:
    cRules m_Rules;
  public:
    cRuleAnd(cRule *Rule1, cRule *Rule2 = NULL, ...) {
      m_Rules.Add(Rule1);
      cRule *next = Rule2;

      va_list ap;
      va_start(ap, Rule2);

      while(next) {
	m_Rules.Add(next);
	next = va_arg(ap, cRule*);
      }
      
      va_end(ap);
    }
    void Add(cRule *Rule) { m_Rules.Add(Rule); }
    virtual bool Matches(cEvent *Event) const {
      bool result = true;
      for(cRule *i = m_Rules.First(); i && result; i=m_Rules.Next(i))
	result = /*result &&*/ i->Matches(Event);
      return result;
    }
    virtual cString ToText(int ident) const { 
      cString str = cString::sprintf("%s{\n%s", Ident(ident), *m_Rules.First()->ToText(ident+4));
      for(cRule *i = m_Rules.Next(m_Rules.First()); i; i=m_Rules.Next(i))
	str = cString::sprintf("%s AND \n%s", *str, *i->ToText(ident+4));
      str = cString::sprintf("%s\n%s}", *str, Ident(ident));
      return str;
    }
};

//--- cRuleOr ---------------------------------------------------

class cRuleOr : public cRule 
{
  protected:
    cRules m_Rules;
  public:
    cRuleOr(cRule *Rule1, cRule *Rule2 = NULL, ...) {
      m_Rules.Add(Rule1);
      cRule *next = Rule2;

      va_list ap;
      va_start(ap, Rule2);

      while(next) {
	m_Rules.Add(next);
	next = va_arg(ap, cRule*);
      }
      
      va_end(ap);
    }
    void Add(cRule *Rule) { m_Rules.Add(Rule); }
    virtual bool Matches(cEvent *Event) const {
      bool result = false;
      for(cRule *i = m_Rules.First(); i && !result; i=m_Rules.Next(i))
	result = /*result ||*/ i->Matches(Event);
      return result;
    }
    virtual cString ToText(int ident) const { 
      cString str = cString::sprintf("%s{\n%s", Ident(ident), *m_Rules.First()->ToText(ident+4));
      for(cRule *i = m_Rules.Next(m_Rules.First()); i; i=m_Rules.Next(i))
	str = cString::sprintf("%s OR\n%s", *str, *i->ToText(ident+4));
      str = cString::sprintf("%s\n%s}\n", *str, Ident(ident));
      return str;
    }
};


//--- cRule -----------------------------------------------------

#define CONFERR(msg...) do { esyslog("autotimer config error: " msg ); return NULL; } while(0)

static void SkipSpace(char **pText)
{
  while(**pText == ' ' || **pText == '\r' || **pText == '\n' || **pText == '\t')
    (*pText)++;
}

static cString TakeString(char **pText)
{
  char str[1024]="", *s = str;
  char *Text = *pText;

  while(*Text && *Text != '\"')
    *Text++ = ' ';
  *Text++ = ' ';

  while(*Text && *Text != '\"') {
    *s++ = *Text;
    *Text++ = ' ';
  }
  *Text = 0;
  *s = 0;

  *pText = Text+1;

  return cString(str);
}

cRule *cRule::RuleFromFile(const char *File)
{
  FILE *f = fopen(File, "r");
  if(f) {
    char str[0x10000+1], *pt = str;
    int n = fread(str, 1, 0x10000, f);
    str[n] = 0;
    cRule *Rule = cRule::Parse(&pt);
    if(!Rule)
      esyslog("autotimer: Error parsing %s", File);
#if 0
    else {
      FILE *fp = fopen("/video/autotimer.spec.parsed", "w");
      fprintf(fp, "%s\n", *Rule->ToText());
      fclose(fp);
    }
#endif
    fclose(f);
    esyslog("autotimer: Rules loaded from %s", File);
    return Rule;
  }
  esyslog("autotimer: Error opening  %s", File);
  return NULL;
}

cRule *cRule::Parse(const char *Text) 
{
  char *tmp = strdup(Text);
  cRule *result = cRule::Parse(&tmp);
  free(tmp);
  return result;
}

cRule *cRule::Parse(char **pText)
{
  cRule *Result = NULL;
  char *Text = *pText;
  cRuleAnd *rand = NULL;
  cRuleOr *ror = NULL;
  //cRuleNot *rnot = NULL;

  while(*Text) {

    /* skip whitespace */
    SkipSpace(pText);
    Text = *pText;
    if(!*Text)
      continue;

    if(!strncmp(Text, "AND", 3)) {
      if(!rand) {
	if(!Result)
	  CONFERR("AND without first rule (%s)", Text);
	rand = new cRuleAnd(Result);
	Result = NULL;
      }
      (*pText) += 3;
    } else if(!strncmp(Text, "OR", 2)) {
      if(!ror) {
	if(!Result)
	  CONFERR("OR without first rule (%s)", Text);
	ror = new cRuleOr(Result);
	Result = NULL;
      }
      (*pText) += 2;
    } else if(!strncmp(Text, "NOT ", 4)) {
      //  *** return new cRuleNot(Parse(Text));
    } else {
      if(Result) 
	CONFERR("multiple rules without operator");

      if(*Text == '{') {
	*(Text++) = 0;
	*pText = Text;
	
	int n = 1, pos = 0, skipping = 0;
	while(Text[pos] && n) {
	  if(skipping && Text[pos] == '\"')
	    skipping = 0;
	  else if(Text[pos] == '\"')
	    skipping = 1;
	  else if(!skipping && Text[pos] == '{')
	    n++;
	  else if(!skipping && Text[pos] == '}')
	    n--;
	  if(n)
	    pos++;
	}
	
	if(!Text[pos] && n)
	  CONFERR("missing closing \'}\'");
	
	Text[pos++] = 0;
	
	Result = Parse(pText);
	*pText = &Text[pos];
      } else if(!strncmp(Text, "TITLE IS", 8)) {
	(*pText) += 8;
	Result = new cRuleTitle(TakeString(pText));
      } else if(!strncmp(Text, "TITLE CONTAINS", 14)) {
	(*pText) += 14;
	Result = new cRuleTitleSubstring(TakeString(pText));
      } else if(!strncmp(Text, "INFO CONTAINS", 13)) {
	(*pText) += 13;
	Result = new cRuleSubstring(TakeString(pText));
      } else if(!strncmp(Text, "CHANNELNUMBER IS", 16)) {
	int s=0, e=0;
	if(2==sscanf(Text+16, "%d-%d", &s, &e))
	  Result = new cRuleChannelRange(s, e);
	else if(1==sscanf(Text+16, "%d", &s))
	  Result = new cRuleChannelRange(s, s);
	else
	  CONFERR("missing channel number");

	(*pText) += 16;
	SkipSpace(pText);
	while(**pText && **pText != ' ')
	  (*pText)++;
      } else if(*Text) {
	CONFERR("unknown rule: %s", Text);
      }
    }

    if(rand && Result) {
      rand->Add(Result);
      Result = NULL;
    } else if(ror && Result) {
      ror->Add(Result);
      Result = NULL;
    }
    if(rand && ror)
      CONFERR("AND and OR mixed");

    Text = *pText;
  }

  if(!Result && rand)
    return rand;
  if(!Result && ror)
    return ror;

  return Result;
}


